/*
 * Copyright (c) 2003-2007 Erez Zadok
 * Copyright (c) 2003-2007 Stony Brook University
 * Copyright (c) 2003-2007 The Research Foundation of State University of New York
 * Copyright (c) 2004-2007 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2007 Gopala Surya
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 * od_async.c
 * output driver for asynchronous output
 */

#include "tracer.h"
#include "od_file.h"
#ifdef KWRAPPERS
#include "kwrappers.h"
#endif
/*
 * internal structure for this driver
 */
struct buffer {
	struct list_head next;
	char *data;
	char *curr;
};

/* async writer thread's main function */
int writerd(void *ptr);

static struct buffer *get_buffer(int size)
{
	struct buffer *tmp = KMALLOC(sizeof(struct buffer), GFP_KERNEL);
	if (!tmp) {
		return NULL;
	}
	tmp->data = vmalloc(size);
	if (!tmp->data) {
		KFREE(tmp);
		return NULL;
	}
	INIT_LIST_HEAD(&tmp->next);
	tmp->curr = tmp->data;
	return tmp;
}

static inline void free_buffer(struct buffer *tmp)
{
	vfree(tmp->data);
	KFREE(tmp);
}

static void free_list(struct list_head *start) {
  struct buffer *entry;
  struct buffer *prev_entry = NULL;
  list_for_each_entry(entry, start, next) {
    if(prev_entry)
      free_buffer(prev_entry);
    prev_entry = entry;
  }
  if(prev_entry)
    free_buffer(prev_entry);
}

static int alloc_list(int len, int size, struct list_head *head)
{
	struct buffer *tmp;
	int i;
	INIT_LIST_HEAD(head);
	if (!len)
		return -EINVAL;
	for (i = 0; i < len; i++) {
		tmp = get_buffer(size);
		if (!tmp) {
			free_list(head);
			return -ENOMEM;
		}
		list_add(&tmp->next, head);
		//printk("tmp is %p\n", tmp);

	}
	return 0;
}


struct od_async_data {
	struct file *file;

	struct list_head full_list;
	spinlock_t full_lock;
	struct semaphore full_sem;

	struct list_head empty_list;
	spinlock_t empty_lock;
	struct semaphore empty_sem;

	struct buffer *curr_buf;
	int buf_size;
	atomic_t ret; /* used to communicate error codes between the two threads */
	atomic_t thread_dead;
	atomic_t thread_die;
	//atomic_t xxx;
};

/**
 * Destroys the file driver object.
 * Make sure that release function has already been called
 * @param output driver to be destroyed
 */

void destroy_async_driver(struct output_driver *od)
{
	ASSERT(od);
	ASSERT(od->od_private_data == NULL);
	/* nothing to do, as of now */
	KFREE(od);
}

/**
 * Initializes the file driver object.
 * @param Async driver object to initialize
 * @param <code>struct od_async_params</code> type initialization parameters
 * @see struct od_async_params
 */

static int async_init(struct output_driver *this, void *params, struct output_driver *next, int level)
{
	int ret = -ENOMEM;
	struct od_async_data *fdata;
	struct file *file = NULL;
	char *name;
	struct od_async_params *fp = (struct od_async_params *)params;
	ASSERT(this->od_private_data == NULL);

	print_entry_location();

	if (!fp && !next) {
		ret = -EINVAL;
		goto out;
	}

	if (next) {
		this->od_next = next;
	} else if (fp->name == NULL) {
		/* File descriptor mode of operation
		 * fget the fd from the current process's
		 * file descriptor table
		 */
		if ((file = fget(fp->fd)) == NULL) {
			ret = -EINVAL;
			goto out;
		}
		if (atomic_read(&file->f_count) != 2) {
			ret = -EBUSY;
			goto out_close;
		}
	} else {

		name = getname(fp->name);
		if (IS_ERR(name)) {
			ret = PTR_ERR(name);
			goto out;
		}
		if (fp->sync_writes)
			file = filp_open(name, O_CREAT | O_WRONLY | O_TRUNC | O_SYNC | O_LARGEFILE, 0644);
		else
			file = filp_open(name, O_CREAT | O_WRONLY | O_TRUNC | O_LARGEFILE, 0644);
		putname(name);
		if (IS_ERR(file)) {
			ret = PTR_ERR(file);
			goto out;
		}
		if (atomic_read(&file->f_count) != 1) {
			ret = -EBUSY;
			goto out_close;
		}
	}

	fdata = this->od_private_data = KMALLOC(sizeof(struct od_async_data), GFP_KERNEL);
	if (!fdata)
		goto out_close;
	memset(this->od_private_data, 0, sizeof(struct od_async_data));
	/* allocate the buffer */
	/* XXX: check for buf_size and buf_count invalidity */
	fdata->buf_size = fp->buf_size;
	INIT_LIST_HEAD(&fdata->empty_list);
	INIT_LIST_HEAD(&fdata->full_list);
	if ((ret = alloc_list(fp->buf_count, fp->buf_size, &fdata->empty_list)) != 0)
		goto out_free;
	ASSERT(!list_empty(&fdata->empty_list));
	/* take one buffer out of the empty buffer list and call it "current" */
	fdata->curr_buf = list_entry(fdata->empty_list.next, struct buffer, next);
	list_del(&fdata->curr_buf->next);
	//printk("curr_buf = %p, curr ptr = %p\n", fdata->curr_buf, fdata->curr_buf->curr);

	/* initialise the semaphores */
	sema_init(&fdata->empty_sem, fp->buf_count - 1);
	sema_init(&fdata->full_sem, 0);

	/* initialize the spin locks */
	spin_lock_init(&fdata->full_lock);
	spin_lock_init(&fdata->empty_lock);

	atomic_set(&fdata->thread_dead, 0);
	atomic_set(&fdata->thread_die, 0);
	//atomic_set(&fdata->xxx, 0);

	fdata->file = file;
	atomic_set(&fdata->ret, 0);

	/* finally, start the kernel thread */
	kernel_thread(writerd, (void *)this, CLONE_FS | CLONE_FILES);
	/* all good! */
	ret = 0;
	goto out;

out_free:
	KFREE(fdata);
	this->od_private_data = NULL;

out_close:
	if (file)
		fput(file);
out:
	print_exit_status(ret);
	return ret;
}

/**
 * Releases the private data.
 * Flush should be called before this. Also causes the chained drivers
 * to be released.
 * @param driver to release
 */
void async_release(struct output_driver *this)
{
	struct od_async_data *fdata;
	ASSERT(this);
	ASSERT(this->od_type == OD_ASYNC);
	fdata = this->od_private_data;
	/* call release on chained driver */
	//if (this->next)
	//	this->next->od_ops->release(this->next);
	/* close the file */
	if (fdata->file)
		fput(fdata->file);

	spin_lock(&fdata->empty_lock);
	spin_lock(&fdata->full_lock);
	ASSERT(list_empty(&fdata->full_list));
	free_list(&fdata->empty_list);
	spin_unlock(&fdata->full_lock);
	spin_unlock(&fdata->empty_lock);

	KFREE(fdata);
	this->od_private_data = NULL;
}

static inline int do_file_write(struct file *file, void *buf, int len)
{
	int ret = 0;
	mm_segment_t oldfs;
	oldfs = set_fs(KERNEL_DS);
	ret = file->f_op->write(file, buf, len, &file->f_pos);
	set_fs(oldfs);
	return ret;
}
/**
 * flushes the interbal buffer, if any.
 * @param object to operate on
 * @return Number of bytes written.
 */
int async_flush(struct output_driver *this)
{
	struct od_async_data *fdata;
	struct buffer *entry;
	struct list_head *pos, *n;
	int ret = 0;
	int tot = 0;

	print_entry_location();
	ASSERT(this);
	ASSERT(this->od_type == OD_ASYNC);
	fdata = this->od_private_data;

	/* kill the kernel thread */
	if (atomic_read(&fdata->ret) == 0) {
		atomic_set(&fdata->thread_die, 1);
		up(&fdata->full_sem);
		while (!atomic_read(&fdata->thread_dead)) {
			schedule();
		}
	}


	/*
	 * now that the kernel thread is dead, we're in control.  We can play around without bothering
	 * for semaphores.
	 */
	/*
	 * if current is empty, add it to the empty list
	 * else, add it to the full list
	 */
	if (fdata->curr_buf->curr == fdata->curr_buf->data)
		list_add_tail(&fdata->curr_buf->next, &fdata->empty_list);
	else
		list_add_tail(&fdata->curr_buf->next, &fdata->full_list);

	list_for_each_safe(pos, n, &fdata->full_list) {
		entry  = list_entry(pos, struct buffer, next);
		if (this->od_next)
			ret = this->od_next->od_ops->write(this->od_next, entry->data, entry->curr - entry->data,OO_KERNEL_BUFFER);
		else
			ret = do_file_write(fdata->file, entry->data, entry->curr - entry->data);
		if (ret != entry->curr - entry->data) {
			tot = ret;
			break;
		}
		tot += ret;
		entry->curr = entry->data;
		list_move_tail(pos, &fdata->empty_list);
	}
	ret = tot;
	if (ret >= 0 && this->od_next)
		ret = this->od_next->od_ops->flush(this->od_next);
	ASSERT(list_empty(&fdata->full_list));
	print_exit_status(ret);
	return ret;
}

/**
 * Writes the data to output.
 * If internal buffering is enabled, data is written to buffer
 * till the buffer runs out.
 * @param object to operate on
 * @param buffer to write out
 * @param number of bytes to write
 * @param whether the buffer is a user-level buffer (OO_USER_BUFFER)
 * @returns number of bytes written
 */

int async_write(struct output_driver *this, void *buf, int len,int is_user_buffer)
{
	int ret = 0;
	int total;
	int remaining;
	struct od_async_data *fdata;
	char* tmp;
	int junk;
	print_entry_location();
	ASSERT(this);
	ASSERT(this->od_type == OD_ASYNC);
	fdata = this->od_private_data;

        tmp = (char *)buf;
	total = len;
	if ((ret = atomic_read(&fdata->ret)) != 0) {
		goto out;
	}
	while(len > 0) {
		remaining = fdata->curr_buf->data + fdata->buf_size - fdata->curr_buf->curr;
		if (remaining > len) {
		        //printk("curr = %p curr ptr = %p\n", fdata->curr_buf, fdata->curr_buf->curr);
		        if(is_user_buffer == OO_KERNEL_BUFFER)
			  memcpy(fdata->curr_buf->curr, tmp, len);
			else
			  junk = copy_from_user(fdata->curr_buf->curr, tmp, len);
			fdata->curr_buf->curr += len;
			len = 0;
		} else {
     		        if(is_user_buffer == OO_KERNEL_BUFFER)
			  memcpy(fdata->curr_buf->curr, tmp, remaining);
			else
			  junk = copy_from_user(fdata->curr_buf->curr, tmp, remaining);
			fdata->curr_buf->curr += remaining;
			len -= remaining;
			tmp += remaining;
			/* throw the buffer to the full list and wake up the other thread */
			spin_lock(&fdata->full_lock);
			list_add_tail(&fdata->curr_buf->next, &fdata->full_list);
			//printk("%d: upping full_sem \n", atomic_read(&fdata->xxx));
			//atomic_inc(&fdata->xxx);
			up(&fdata->full_sem);
			spin_unlock(&fdata->full_lock);

			/* get the next buffer */
			//printk("%d: downing empty_sem \n", atomic_read(&fdata->xxx));
			//atomic_inc(&fdata->xxx);
			down(&fdata->empty_sem);
			//printk("%d: downed empty_sem \n", atomic_read(&fdata->xxx));
			//atomic_inc(&fdata->xxx);
			spin_lock(&fdata->empty_lock);
			fdata->curr_buf = list_entry(fdata->empty_list.next, struct buffer, next);
			list_del(&fdata->curr_buf->next);
			spin_unlock(&fdata->empty_lock);
		}
	}
	ret = total;

out:
	print_exit_status(ret);
	return ret;
}

/**
 * Returns the preferred block size
 * @param object to operate on
 * @returns block size
 */
static int async_getblocksize(struct output_driver *this)
{
    struct od_async_data *fdata;
    ASSERT(this);
    ASSERT(this->od_type == OD_ASYNC);
    fdata = this->od_private_data;
    return fdata->buf_size;
}

static struct output_ops async_od_ops = {
    write: async_write,
    flush: async_flush,
    getblocksize: async_getblocksize,
    init: async_init,
    release: async_release
};

/**
 * Creates a file driver object.
 * @return A file output driver object with function pointers initialised
 */

struct output_driver* create_async_driver(void)
{
    struct output_driver *driver = (struct output_driver *)KMALLOC(sizeof(struct output_driver), GFP_KERNEL);

    if (driver) {
	driver->od_ops = &async_od_ops;
	driver->od_type = OD_ASYNC;
	driver->od_private_data = NULL;
	driver->od_next = NULL;
    }
    return driver;
}

/*
 * Kernel thread for writing out the data asynchronously
 */
int writerd(void *ptr)
{
	struct output_driver *this = (struct output_driver *)ptr;
	struct od_async_data *fdata = (struct od_async_data *)(this->od_private_data);
	struct output_ops *ops = NULL;

	struct buffer *curr_buf;
	int ret;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	daemonize("ktracefsd");
#else
	daemonize();
#endif
	sprintf(current->comm, "tracefswriterd");
	if (this->od_next) {
		ops = this->od_next->od_ops;
	}

	while (1) {
		//printk("%d: downing full_sem \n", atomic_read(&fdata->xxx));
		//	atomic_inc(&fdata->xxx);
		down(&fdata->full_sem);
		//printk("%d: downed full_sem \n", atomic_read(&fdata->xxx));
		//	atomic_inc(&fdata->xxx);
		if (atomic_read(&fdata->thread_die))
			break;
		spin_lock(&fdata->full_lock);
		curr_buf = list_entry(fdata->full_list.next, struct buffer, next);
		list_del(&curr_buf->next);
		spin_unlock(&fdata->full_lock);

		if (ops)
			ret = ops->write(this->od_next, curr_buf->data, curr_buf->curr - curr_buf->data,OO_KERNEL_BUFFER);
		else
			ret = do_file_write(fdata->file, curr_buf->data, curr_buf->curr - curr_buf->data);
		curr_buf->curr = curr_buf->data;
		spin_lock(&fdata->empty_lock);
		list_add_tail(&curr_buf->next, &fdata->empty_list);
		//printk("%d: upping empty_sem \n", atomic_read(&fdata->xxx));
		//	atomic_inc(&fdata->xxx);
		up(&fdata->empty_sem);
		spin_unlock(&fdata->empty_lock);
		if (ret < 0) {
			atomic_set(&fdata->ret, ret);
			break;
		}

	}
	atomic_set(&fdata->thread_dead, 1);

	return 0;

}
