/*
 * Copyright (c) 2003-2007 Erez Zadok
 * Copyright (c) 2003-2007 Stony Brook University
 * Copyright (c) 2003-2007 The Research Foundation of State University of New York
 * Copyright (c) 2004-2007 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2007 Gopala Surya
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 * od_encrypt.c
 * output driver for encryption
 */

#include "ciphers/crypto.h"
#include "tracer.h"
#include "od_encrypt.h"
#include "ciphers/gen-cipher.h"
#ifdef KWRAPPERS
#include "kwrappers.h"
#endif
//extern struct tf_cipher_implementation blowfish_cbc;

/*
 * internal structure for this driver
 */

struct od_encrypt_data {

	char *buf;     /* buffer */
	int size; /* size of the buffer */
	char *begin; /* start of unencrypted data */
	char *end;  /* end of unencrypted data */
	struct tf_cipher_implementation *ci;
	struct tf_cipher_context *cx;

};

/**
 * Destroys the checksum driver object.
 * Make sure that release function has already been called
 * @param output driver to be destroyed
 */

void destroy_encrypt_driver(struct output_driver *od)
{
	ASSERT(od);
	ASSERT(od->od_private_data == NULL);
	/* nothing to do, as of now */
	KFREE(od);
}

/**
 * Initializes the encryption driver object.
 * @param Encryption driver object to initialize
 * @param <code>struct od_encrypt_params</code> type initialization parameters
 * @see struct od_encrypt_params
 */

static int encrypt_init(struct output_driver *this, void *params, struct output_driver *next, int level)
{
	int ret = 0;
	struct od_encrypt_data *edata;
	char *key = NULL;
	struct od_encrypt_params *ep = (struct od_encrypt_params *)params;
	ASSERT(this->od_private_data == NULL);
	print_entry_location();
	if (!next) {
		ret = -EINVAL;
		goto out;
	}


	edata = KMALLOC(sizeof(struct od_encrypt_data), GFP_KERNEL);
	if (!edata) {
		ret = -ENOMEM;
		goto out;
	}
	memset(edata, 0, sizeof(*edata));
	switch (ep->algorithm) {
		case ENCRYPT_BLOWFISH:
			edata->ci = &blowfish_cbc;
			break;
		default:
			ret = -EINVAL;
			goto out_free;
			break;
	}


	if ((key = KMALLOC(ep->keylen, GFP_KERNEL)) == NULL) {
		ret = -ENOMEM;
		goto out_free;
	}
	if (copy_from_user(key, ep->key, ep->keylen)) {
		ret = -EFAULT;
		goto out_free;
	}
/*
	{
		int i;
		printk("key = ");
		for (i = 0; i < ep->keylen; i++)
			printk("%x", ep->key[i]);
		printk("\n");
	}
*/
	if ((ret = tf_register_cipher(edata->ci)) != 0) {
		goto out_free;
	}
	edata->size = next->od_ops->getblocksize(next);
	// round of size to block size. This should usually result
        // in a buffer the same size as lower layer's preferred size
        // since buffers should usually be a nice multiple of 8 bytes
	edata->size = edata->size + (edata->size % edata->ci->blocksize);
	/*
         * allocate the ciphertext buffer, keep an extra block size
         * space just in case it's needed for padding.
         * This will not show up in the size and used only for flush
         */
	if ((edata->buf = vmalloc(edata->size + edata->ci->blocksize)) == NULL) {
		ret = -ENOMEM;
		goto out_free;
	}
	edata->begin = edata->end = edata->buf;

	edata->cx = edata->ci->realloc_context(NULL, edata->ci, ep->keylen);
	if (!edata->cx) {
		ret = -EINVAL;
		goto out_free;
	}
	if (edata->ci->set_key(edata->cx, key, ep->keylen) < 0) {
		edata->ci->wipe_context(edata->cx);
		edata->ci->free_context(edata->cx);
		ret = -EINVAL;
		goto out_free;
	}
	KFREE(key);
	key = NULL;
	memset(edata->cx->iv, 0, edata->ci->ivsize);
	/* all good */
	this->od_private_data = edata;
	this->od_next = next;
	goto out;

out_free:
	if (key)
		KFREE(key);
	if (edata->buf)
		vfree(edata->buf);
	KFREE(edata);

out:
	print_exit_status(ret);
	return ret;
}

/**
 * Releases the private data.
 * Flush should be called before this. Also causes the chained drivers
 * to be released.
 * @param driver to release
 */
void encrypt_release(struct output_driver *this)
{
	struct od_encrypt_data *edata;
	ASSERT(this);
	ASSERT(this->od_type == OD_ENCRYPT);
	edata = (struct od_encrypt_data *)this->od_private_data;
	edata->ci->wipe_context(edata->cx);
	edata->ci->free_context(edata->cx);
	tf_unregister_cipher(edata->ci);
	vfree(edata->buf);
	KFREE(edata);
	this->od_private_data = NULL;
}

/**
 * flushes the internal buffer, if any.
 * @param object to operate on
 * @return Number of bytes written.
 */
int encrypt_flush(struct output_driver *this)
{
	struct od_encrypt_data *edata = (struct od_encrypt_data *)this->od_private_data;
	int ret = 0;
	int pad = edata->ci->blocksize - (edata->end - edata->begin) % edata->ci->blocksize;

	print_entry_location();
	ASSERT(this);
	ASSERT(this->od_type == OD_ENCRYPT);
	ASSERT(this->od_next);

	pkcspad_inplace(edata->buf, edata->end - edata->buf + pad, edata->end - edata->buf);
	edata->end += pad;
	edata->ci->encrypt(edata->cx, edata->begin, edata->begin, edata->end - edata->begin);

	this->od_next->od_ops->write(this->od_next, edata->buf, edata->end - edata->buf,OO_KERNEL_BUFFER);
	edata->begin = edata->end = edata->buf;

	ret = this->od_next->od_ops->flush(this->od_next);

	print_exit_status(ret);
	return ret;
}

int encrypt_write(struct output_driver *this, void *buf, int len,int is_user_buffer)
{
	struct od_encrypt_data *edata = (struct od_encrypt_data *)this->od_private_data;
	int ret = 0;
	int diff = (len + edata->end - edata->begin) % edata->ci->blocksize;
	int new_bytes = len - diff;
	int remaining;
	int junk;
	char *curr = buf;
	print_entry_location();
	ASSERT(this);
	ASSERT(this->od_type == OD_ENCRYPT);
	/* Encryption driver cannot work in isolation */
	ASSERT(this->od_next);
	remaining = edata->buf + edata->size - edata->end;
	while (new_bytes > remaining) {
	        if(is_user_buffer == OO_KERNEL_BUFFER)
		  memcpy(edata->end, curr, remaining);
		else
		  junk = copy_from_user(edata->end, curr, remaining);
		edata->end += remaining;
		edata->ci->encrypt(edata->cx, edata->begin, edata->begin, edata->end - edata->begin);
		edata->begin = edata->end;
		ret = this->od_next->od_ops->write(this->od_next, edata->buf, edata->begin - edata->buf,is_user_buffer);
		if (ret < 0) {
			edata->begin = edata->end = edata->buf;
			goto out;
		}
		edata->begin = edata->end = edata->buf;
		curr += remaining;
		new_bytes -= remaining;
		remaining = edata->size;
	}
	if (new_bytes) {
 	        if(is_user_buffer == OO_KERNEL_BUFFER)
		  memcpy(edata->end, curr, new_bytes);
		else
		  junk = copy_from_user(edata->end, curr, new_bytes);
		edata->end += new_bytes;
		edata->ci->encrypt(edata->cx, edata->begin, edata->begin, edata->end - edata->begin);
		edata->begin = edata->end;
	}

	if (diff != 0) {
		if (edata->buf + edata->size - edata->end < diff) {
			/* no space for the new characters, flush them out */
			ret = this->od_next->od_ops->write(this->od_next, edata->buf, edata->begin - edata->buf,is_user_buffer);
			if (ret < 0) {
				edata->begin = edata->end = edata->buf;
				goto out;
			}
			edata->begin = edata->end = edata->buf;
		}
		if(is_user_buffer == OO_KERNEL_BUFFER)
		  memcpy(edata->end, buf + len - diff, diff);
		else
		  junk = copy_from_user(edata->end, buf + len - diff, diff);
		edata->end += diff;
	}

out:
	ret = (ret >= 0) ? len : ret;
	print_exit_status(ret);
	return ret;
}

/**
 * Returns the preferred block size
 * @param object to operate on
 * @returns block size
 */
int encrypt_getblocksize(struct output_driver *this)
{
	ASSERT(this);
	ASSERT(this->od_type == OD_ENCRYPT);
	return PAGE_SIZE;
}

static struct output_ops encrypt_od_ops = {
write: encrypt_write,
flush: encrypt_flush,
getblocksize: encrypt_getblocksize,
init: encrypt_init,
release: encrypt_release
};

/**
 * Creates a file driver object.
 * @return A file output driver object with function pointers initialised
 */

struct output_driver* create_encrypt_driver()
{
	struct output_driver *driver = (struct output_driver *)KMALLOC(sizeof(struct output_driver), GFP_KERNEL);

	if (driver) {
		driver->od_ops = &encrypt_od_ops;
		driver->od_type = OD_ENCRYPT;
		driver->od_private_data = NULL;
		driver->od_next = NULL;
	}
	return driver;
}


